'Filename   Millivoltmeter_Mega328P_110.bas
'Version    1.10 Arduino NANO
'Date       2018-01-25
'Purpose    Digital display for RMS Millivoltmeter
'           Modes of operation:
'           bytMode  = 1 Voltage measurement
'                    = 2 Voltage  level  measurement (dBu, ref. 775mV)
'                    = 3 Voltage  level  measurement (dBV, ref. 1V)
'                    = 4 Absolute power  measurement (dBm, ref. 775mV@600Ohm)
'                    = 5 Relative level  measurement (dB, free choice ref.)
'                    = 6 Setup
'           A 10:1 probe may be used except in mode 4 (dBm power)
'Author     Georg Latzel, DL6GL
'Copyright  (c) Georg Latzel, free for private use under GNU license
'Hardware   Arduino NANO (ATmega328p)
'           Crystal 16 MHz
'           Port  PB0(D 8):   Mode key
'           Port  PB1(D 9):   OK   key
'           Port  PB2(D10):   Up   key
'           Port  PB3(D11):   Down key
'           Port  PC0(A 0):   Range select  200mV
'           Port  PC1(A 1):   Range select    2 V
'           Port  PC2(A 2):   Range select   20 V
'           Port  PC3(A 3):   Range select  200 V
'           Port  PC4(A 4):   SDA to ADC
'           Port  PC5(A 5):   SCL to ADC
'           Port  PD0(RX0):   Range select    2mV
'           Port  PD1(TX1):   Range select   20mV
'           Ports PD2 ... PD7 (D2...D7): 2x16 Text LCD
'Compiler   BASCOM-AVR 2.0.7.6
'           For readable "camel case" variables mark "Don't change case"
'           in BASCOM Options-Programmer-Environment-Editor
'Edit history
'V1.00      2017-12-01 First release
'           Flash used: 13,420 bytes (40%)
'V1.10      2018-01-25
'           Added: Mode 6, Relative level  measurement (%, free choice ref.)
'           Setup is mode 7 now.
'           Flash used: 14,076 bytes (42%)

'==============================================================================
'FUSE-Bits (AVR Studio 4.0-Settings, [ ]=no check, [x]=check)
'In ATMEGA-Data sheets: Bit = 1 -> not set = no check
'                       Bit = 0 -> set     = check
'BODLEVEL   Brown-out detection disabled
'RSTDISBL   [ ]
'DWEN       [ ]
'SPIEN      [x] Do not change!
'WDTON      [ ]
'EESAVE     [x]
'BOOTSZ     Boot Flash size=256 words (minimal boot size)
'BOOTRST    [ ]
'CKDIV8     [ ]
'CKOUT      [ ]
'SUT_CKSEL  Ext. Crystal Osc. 8.0- MHz, Start-up time 16k + 14ms
'Extended:  0xFF
'High:      0xD7
'Low:       0xFF

$regfile = "m328pdef.dat"                         'ATMega328P
$crystal = 16000000                               '16 MHz xtal
$hwstack = 64
$swstack = 64
$framesize = 64

'Configure in/out on Port B ---------------------------------------------------
'Key Mode                  D8  PB0(IC01)
'Key OK                    D9  PB1(OC1A)
'Key Up                    D10 PB2(OC1B)
'Key Down & ISP MOSI       D11 PB3(MOSI)
'ISP MISO                  D12 PB4(MISO)
'ISP SCK                   D13 PB5(SCK)
'Up and Down keys with autorepeat to set data
'OK key pressed short:
'- Voltage & voltage level : 10:1 probe is used
'- Setup                   : next setup submenu , current data are not changed
'OK key pressed long :
'- Relative level          : Set current level as offset
'- Setup                   : store current setting to EEPROM
DDRB = &B00000000                                 'All input
PortB = &B00001111                                'Pullup PB0...PB3
Const Key_Mode = 0                                'Key Mode        @ PinB.0
Const Key_OK = 1                                  'Key OK          @ PinB.1
Const Key_Up = 2                                  'Key Up          @ PINB.2
Const Key_Down = 3                                'Key Down        @ PinB.3
Key_port Alias PinB                               'Input-Port Keys
Dim bytKey_Pressed As Byte                        '>0 if any key  is pressed
Dim bytKey_Up As Byte                             '=1 if key up   is pressed
Dim bytKey_Down As Byte                           '=1 if key down is pressed
Dim bytKey_Mode As Byte                           '=1 if Key mode is pressed
Dim bytKey_OK As Byte                             'Key OK  pressed: 1(short), 2(long)
Dim bytKey_mask As Byte                           'Mask for short/long keys
bytKey_Mask = &B00001111                          'Set 1 to enable key bits 0..3

'Configure in/out on Port C ---------------------------------------------------
'Enable range 3            A0  PC0/ADC0
'Enable range 4            A1  PC1/ADC1
'Enable range 5            A2  PC2/ADC2
'Enable range 6            A3  PC3/ADC3
'I2C SDA                   A4  PC4/ADC4/SDA
'I2C SCL                   A5  PC5/ADC5/SCL
'Not used                  A6  ADC6
'Not used                  A7  ADC7
DDRC = &B00001111                                 'PC0...PC3 output, PC4...PC7 input
PORTC = &B00000000
CONFIG SDA = PORTC.4                              'I2C SDA
CONFIG SCL = PORTC.5                              'I2C SCL
RngPort2 Alias PortC

'Configure LCD on Port D ------------------------------------------------------
'LCD RS                    D7  PD7
'LCD E                     D6  PD6
'LCD D4                    D5  PD5
'LCD D5                    D4  PD4
'LCD D6                    D3  PD3
'LCD D7                    D2  PD2
'Enable range 2            TX1 PD1
'Enable range 1            RX0 PD0
Config Lcd = 16 * 2
Config Lcdpin = Pin , Db7 = PortD.2 , Db6 = PortD.3 , Db5 = PortD.4
Config Lcdpin = Pin , Db4 = PortD.5 , E = PortD.6 , Rs = PortD.7
Config Lcdbus = 4
Const bytChar = 16                                'Characters per LCD line
DDRD.0 = 1                                        'PD0 output
DDRD.1 = 1                                        'PD1 output
RngPort1 Alias PortD

'Variables program control ----------------------------------------------------
Dim bitSignal As Bit                              'Text toggle
Dim bytTimer0_Signal As Byte                      'Signal Timer0 counter
Dim bytRMSErr As Byte                             'AD736 RMS input indicator
Dim bytMode As Byte
Dim bytInitMenu As Byte                           '=0: enter setup, =2: change value
Dim bytMenu As Byte                               'Setup submenu number
Dim bytRange As Byte                              'Meter range 1...7, 7; 200V with 10:1 probe
Dim bytRangeMax As Byte                           'Max range =6, with 10:1 probe =7
Dim bytRangeMin As Byte                           'Min range =1, with 10:1 probe =2
Dim bytDecimals As Byte                           'Decimals to display
Dim bytRelSet As Byte                             '=1: Relative level is set
Dim bytProbeSet As Byte                           '=1: 10:1 probe is used
Dim bytProbeCur As Byte                           'Current probe status
Dim bytCursorFirst As Byte                        'First LCD cursor position
Dim strRanges(7) As String * 6                    'Meter ranges

'Common Variables, multiply used ----------------------------------------------
Dim bytData As Byte                               'Handled in Sub SetValue
Dim bytDataMin As Byte                            'Handled in Sub SetValue
Dim bytDataMax As Byte                            'Handled in Sub SetValue
Dim bytDataInc As Byte                            'Handled in Sub SetValue
Dim bytDataDec As Byte                            'Handled in Sub SetValue
Dim wrdData As Word                               'Handled in Sub SetValue
Dim wrdDataMin As Word                            'Handled in Sub SetValue
Dim wrdDataMax As Word                            'Handled in Sub SetValue
Dim strUnit As String * 3                         'Handled in Sub SetValue
Dim bytTmp0 As Byte
Dim bytTmp1 As Byte
Dim intTmp0 As Integer
Dim dwdTmp0 As DWord
Dim sngTmp0 As Single
Dim sngTmp1 As Single
Dim dblTmp0 As Double
Dim strDis1 As String * 16                        'Text to be dispayed on LCD
Dim strDis2 As String * 16
Dim strTmp16 As String * 16

'Variables Keys ---------------------------------------------------------------
Dim bytIwr0 As Byte
Dim bytKey_ct0 As Byte
Dim bytKey_ct1 As Byte
Dim bytKey_state As Byte
Dim bytKey_press As Byte
Dim bytKey_save As Byte
Dim bytTimer0_keys As Word                        'Timer0 counter for keys
Const bytTimer0_long = 80                         'Timer0 counts for long press

'External MCP3421 ADC Variables -----------------------------------------------
Dim bytADCConf As Byte                            'ADC conversion register byte
Dim bytADCRes As Byte                             'ADC resolution 16 or 18bits
Dim bytADCRep As Byte                             'Repetitive ADC runs 1...10
Dim bytADCRun As Byte                             'Current ADC run for mean value
Dim bytADCGain As Byte                            'ADC gain 1/2/4/8
Dim bytADCOvl As Byte                             '=1: ADC overflow, =0: OK
Dim bytADCResult As Byte                          '=1: ADC result avalilable
Dim wrdADCRef As Word                             'ADC reference voltage (mVx10)
Const bytADCRepMax = 10                           'Max ADC repetitions
Dim lngADCVal As Long                             '18bit ADC conversion result
Dim lngADCArr(bytADCRepMax) As Long               'ADC results of repetitive runs
Dim bytADCVal(4) As Byte At lngADCVal Overlay
Dim dwdADCSum As DWord                            'Sum of ADC repetitive runs
Dim dwdADCMean As DWord                           'Mean ADC conversion result
Dim dwdADCOvl As DWord                            'Max. ADC count for chosen resolution
Dim dwdADCOffset As DWord                         '2mV offset
Dim sngVolt As Single                             'Measured converted voltage
Dim sngVoltRef As Single                          'Reference voltage for rel. level
Dim dblADCCalFac As Double                        'Calibration factor 18Bit (mV)                 '
Const bytADCRead = &B11010001                     'ADC read address, '1101'=device code
Const bytADCWrite = &B11010000                    'ADC write address

'EEPROM data ------------------------------------------------------------------
Dim bytEEPStored As Byte                          '=1: Setup data stored to EEPROM
Dim wrdAddr As Word
Dim wrdEEPROM As Word                             'Word variable to/from EEPROM
Dim bytWrdEEP(2) As Byte At wrdEEPROM Overlay
Dim dwdEEPROM As DWord                            'DWord variable to/from EEPROM
Dim bytDwdEEP(4) As Byte At dwdEEPROM Overlay

'Declare subs & functions -----------------------------------------------------
'Display subs
Declare Sub ShowRange
Declare Sub Probe
Declare Sub ClearLCDLine(byVal bytRow As Byte , byVal bytChars As Byte)
Declare Sub ShowVoltage
Declare Sub SetValue(byVal bytRow As Byte , byVal bytType As Byte , byVal wrdScale As Word)
'Key subs
Declare Function KeyPressedLong(byVal bytKey_Long As Byte) As Byte
Declare Function KeyAutoRepeat(byVal bytKey_Rep As Byte) As Byte
'External ADC
Declare Sub GetADCMean
Declare Function ADCCheck() As Byte
Declare Sub ADCSetGain
Declare Sub ADCSetCal
Declare Sub GetVoltage
'EEPROM & EEPROM setup subs
Declare Sub CheckEEPROM
Declare Sub EEPROMReadWrd(byRef wrdVal As Word)
Declare Sub EEPROMWriteWrd(byRef wrdVal As Word)
Declare Sub EEPROMReadDwd(byRef dwdVar As DWord)
Declare Sub EEPROMWriteDwd(byRef dwdVar As DWord)

'Configure Timer0-Interrupt (Timer0 = 8bit, 2^8=256 counts) -------------------
'Possible Prescales: 8, 64, 256, 1024, here 1024
'For 16.000 MHz xtal:
'Timer0 overflow counts are reduced by PresetTimer0 = 56 to 256-56=200
'Overflow time: 200 * 1024 / 16,000,000 = 12.8 msec (78.125 Hz)
Config Timer0 = Timer , Prescale = 1024
Const Presettimer0 = 56
On Timer0 Timer0_isr                              'On Overflow go to Timer0_isr

'Initialize Key Data ----------------------------------------------------------
bytKey_ct0 = 0
bytKey_ct1 = 0
Decr bytKey_ct0
bytKey_ct1 = bytKey_ct0
bytKey_state = 0
bytKey_press = 0

'Start timer ------------------------------------------------------------------
Enable Timer0                                     'Timer0 for keys & PEP timing
Enable Interrupts                                 'enable all Interrupts
Timer0 = Presettimer0                             'initialize Timer0
bytTimer0_keys = 0

'Splash screen ----------------------------------------------------------------
Cls
Cursor Off
Locate 1 , 1
LCD "RMS Meter V 1.10"
Locate 2 , 1
LCD "DL6GL 2018-012-25"
Wait 1
Cls

'Initialize data --------------------------------------------------------------
strRanges(1) = "  2 mV"                           'Voltage ranges
strRanges(2) = " 20 mV"
strRanges(3) = "200 mV"
strRanges(4) = "   2 V"
strRanges(5) = "  20 V"
strRanges(6) = " 200 V"
strRanges(7) = "2000 V"                           'Only for 10:1 probe at 200V
bitSignal = 0
bytMode = 1                                       'Voltage measurement
bytRange = 6                                      'Range 6 (200V)
bytRangeMin = 1
bytRangeMax = 6
bytDecimals = 1
RngPort2.3 = 1                                    'Enable range 6
RngPort1.0 = 1                                    'Disable range 1 (neg. logic)
RngPort1.1 = 1                                    'Disable range 2 (neg. logic)
bytInitMenu = 0
bytRelSet = 0                                     'Relative level not set
bytProbeSet = 0                                   '10:1 probe not used
bytProbeCur = 0

Call CheckEEPROM                                  'Read/initialize EEPROM data
Cls

'Check I2C function of ADC ----------------------------------------------------
Locate 1 , 1
LCD "Check ADC-I2C..."
If ADCCheck() = 0 Then                            'Try to write write address
   Locate 2 , 1                                   '...no success
   LCD "... is not ready"
   Stop                                           'That's the end
Else
   Locate 2 , 1
   LCD "... is ready    "
   Wait 1
End If

'Configure MCP3421 ADC to Gain 1/18bit/continuous mode ------------------------
bytADCGain = 1                                    'ADC programmable gain (PGA)
bytADCRes = 18                                    '18 bit ADC resolution (3.75 SPS)
dwdADCOvl = 131071                                '18 bit ADC overflow = 2^17 - 1
Call ADCSetGain                                   'Set gain=1 & 18bit resolution
Call ADCSetCal                                    'Set calibration factor 18bit

'Initialize data for external ADC running average -----------------------------
dwdADCMean = 0                                    'ADC mean value
dwdADCSum = 0                                     'ADC conversion sum
bytADCRun = 1                                     'ADC result array index
For bytTmp0 = 1 To bytADCRepMax
   lngADCArr(bytTmp0) = 0
Next bytTmp0

Do

   'Check keys, any key pressed? ----------------------------------------------
   'bytKey_press returned from Timer0_isr
   'Key_Mode:  #0
   'Key_OK:    #1
   'Key_Up:    #2
   'Key_Down:  #3
   'Just pressed detected in Timer0_isr: pressed status flag bit = 1 bytKey_press.x
   '2 key flags are issued from Timer0_isr:
   '-1  bytKey_press.x = 1: key x changed from open to pressed,
   '    Used for normal key pressed action
   '    After evaluating bytKey_press.x flag is reset just below.
   '    Short/long pressed status flag is reset in function KeyPressedLong
   '-2  bytKey_state.x = 1: while key is pressed, = 0: Key is released
   '    bytKey_state.x flag is used for autorepeat
   'Key action is stored in the variables:
   'bytKey_Up     = 0: Key Up was not pressed
   '              = 1: Key Up was pressed short
   '              = 2: Key Up was pressed long, auto repeat
   'bytKey_Down   = 0: Key Down was not pressed
   '              = 1: Key Down was pressed short
   '              = 2: Key Down was pressed long, auto repeat
   'bytKey_OK     = 0: Key OK   was not pressed
   '              = 1: Key OK   was pressed short
   '              = 2: Key OK   was pressed long
   'bytKey_Mode   = 0: Key Mode was not pressed
   '              = 1: Key Mode was pressed

   bytKey_Pressed = bytKey_press AND bytKey_mask  'Mask out keys

   If bytKey_Pressed > 0 Then                     'Any key was pressed
      'Check key Mode
      If bytKey_press.Key_Mode = 1 Then           'Key Mode was pressed
         bytKey_Mode = KeyPressedLong(Key_Mode)   'Pressed short or long
      Else
         bytKey_Mode = 0                          'Key Mode not pressed
         bytKey_press.Key_Mode = 0                'Reset Key Mode pressed flag
      End If
      'Check short/long key OK
      If bytKey_press.Key_OK = 1 Then             'Key OK was pressed
         bytKey_OK = KeyPressedLong(Key_OK)       'Pressed short or long
      Else
         bytKey_OK = 0                            'Key OK not pressed
         bytKey_press.Key_OK = 0                  'Reset Key OK pressed flag
      End If
      'Check auto repeat keys Up & Down
      If bytKey_press.Key_Up = 1 Then             'Key Up was pressed
         bytkey_Up = KeyAutoRepeat(Key_Up)        'Pressed short or auto repeat
      Else
         bytKey_state.Key_Up = 0                  'Reset Key Up status  flag
      End If
      If bytKey_press.Key_Down = 1 Then           'Key Down was pressed
         bytkey_Down = KeyAutoRepeat(Key_Down)    'Pressed short or auto repeat
      Else
         bytKey_state.Key_Down = 0                'Reset Key Down status  flag
      End If
   Else                                           'No key was pressed
      bytKey_Up = 0                               'Reset all key flags
      bytKey_Down = 0
      bytKey_OK = 0
      bytKey_Mode = 0
      bytKey_Pressed = 0
   End If

   'Change mode ---------------------------------------------------------------

   If bytKey_Mode > 0 Then                        'Mode key pressed
      Incr bytMode                                'Next Mode
      If bytMode = 4 Then                         'Absolute power measurement
         bytProbeSet = 0                          'No 10:1 probe
      Else
         bytProbeSet = bytProbeCur                'Saved probe status
      End If
      If bytMode > 7 Then                         'Was the last
         bytMode = 1
      End If
      bytInitMenu = 0                             'Initialize menu
      bytMenu = 10 * bytMode
      bytMenu = bytMenu + 1
   End If

   'Change range --------------------------------------------------------------
   If bytMode < 6 Then                            'All modes except setup
      If bytKey_Up = 1 Or bytKey_Down = 1 Then    'Change range
         If bytKey_Up = 1 Then                    'Key up pressed, increase range
            If bytRange < bytRangeMax Then Incr bytRange
         End If
         If bytKey_Down = 1 Then                  'Key down pressed, decrease range
            If bytRange > bytRangeMin Then Decr bytRange
         End If
         Call ShowRange
         'Clear range setting ports PD0, PD1, PC0...PC3 first
         For bytTmp0 = 0 To 1                     'Clear range enable PD0...PD1
            RngPort1.bytTmp0 = 1                  'Negative logic
         Next bytTmp0
         For bytTmp0 = 0 To 3                     'Clear range enable PC0...PC3
            RngPort2.bytTmp0 = 0                  'Positive logic
         Next bytTmp0
         'Now set range setting port for selected range
         If bytProbeSet = 0 Then                  'No 10:1 probe
            bytTmp1 = bytRange                    'Normal range 1...6
         Else                                     '10:1 probe is active
            bytTmp1 = bytRange - 1                'Increased range 2...7
         End If
         If bytTmp1 < 3 Then                      'Range enable PD0...PD1
            bytTmp0 = bytTmp1 - 1
            RngPort1.bytTmp0 = 0                  'Enable range (neg. logic)
         Else                                     'Range enable PC0...PC3
            bytTmp0 = bytTmp1 - 3
            RngPort2.bytTmp0 = 1                  'Enable range (pos. logic)
         End If
      End If
   End If

   Select Case bytMode
      Case 1
         '=====================================================================
         'Voltage measurement
         '=====================================================================
         If bytInitMenu = 0 Then                  'Prepare first line
            Cls
            Call ShowRange
            bytInitMenu = 1
         End If

         Call GetVoltage                          'Voltage from ADC acquisition
         Call ShowVoltage

         If bytKey_OK = 1 Then                    'Key OK pressed short
            Call Probe                            'Enable or disable 10:1 probe
         End If

      Case 2
         '=====================================================================
         'Voltage level dBu, reference 775mV
         '=====================================================================
         If bytInitMenu = 0 Then                  'Prepare first line
            Cls
            Call ShowRange
            bytInitMenu = 1
         End If

         Call GetVoltage                          'Voltage from ADC acquisition
         If bytRange < 4 Then                     'Voltage in mV
            sngTmp0 = sngVolt / 774.6
         Else                                     'Voltage in V
            sngTmp0 = sngVolt / 0.7746
         End If
         Locate 2 , 1
         If bytRMSErr < 2 Then                    'AD736 input < 900mV
            sngTmp0 = LOG10(sngTmp0)
            sngTmp0 = 20 * sngTmp0                'dBu (0.775V = 0dBu)
            sngTmp0 = 100 * sngTmp0               'Show 2 decimals
            sngTmp0 = sngTmp0 + 0.5               'Rounding of last digt
            intTmp0 = sngTmp0
            strDis2 = str(intTmp0)
            strDis2 = Format(strDis2 , "+ 0.00")
         Else                                     'AD736 overrange
            If bitSignal = 1 Then
               strDis2 = "OvRng  "
            Else
               strDis2 = "      "
            End If
         End If
         LCD strDis2

         If bytKey_OK = 1 Then                    'Key OK pressed short
            Call Probe                            'Enable or disable 10:1 probe
         End If

      Case 3
         '=====================================================================
         'Voltage level dBV, reference 1V
         '=====================================================================
         If bytInitMenu = 0 Then                  'Prepare first line
            Cls
            Call ShowRange
            bytInitMenu = 1
         End If

         Call GetVoltage                          'Voltage from ADC acquisition
         If bytRange < 4 Then                     'Voltage in mV
            sngTmp0 = sngVolt * 0.001
         Else                                     'Voltage in V
            sngTmp0 = sngVolt
         End If
         Locate 2 , 1
         If bytRMSErr < 2 Then                    'AD736 input < 900mV
            sngTmp0 = LOG10(sngTmp0)
            sngTmp0 = 20 * sngTmp0                'dBV (1V = 0dBV)
            sngTmp0 = 100 * sngTmp0               'Show 2 decimals
            sngTmp0 = sngTmp0 + 0.5               'Rounding of last digt
            intTmp0 = sngTmp0
            strDis2 = str(intTmp0)
            strDis2 = Format(strDis2 , "+ 0.00")
         Else                                     'AD736 overrange
            If bitSignal = 1 Then
               strDis2 = "OvRng  "
            Else
               strDis2 = "      "
            End If
         End If
         LCD strDis2

         If bytKey_OK = 1 Then                    'Key OK pressed short
            Call Probe                            'Enable or disable 10:1 probe
         End If

      Case 4
         '=====================================================================
         'Absolute power measurement, reference 775mV @ 600 Ohm
         '=====================================================================
         If bytInitMenu = 0 Then                  'Prepare first line
            Cls
            Locate 1 , 1
            LCD "Input impedance"
            Locate 2 , 1
            LCD "***  600" ; chr(244) ; "  ***!"
            Wait 2
            Cls
            Call ShowRange
            bytInitMenu = 1
         End If

         Call GetVoltage                          'Voltage from ADC acquisition
         If bytRange < 4 Then                     'Voltage in mV
            sngTmp0 = sngVolt / 774.6
         Else                                     'Voltage in V
            sngTmp0 = sngVolt / 0.7746
         End If
         Locate 2 , 1
         If bytRMSErr < 2 Then                    'AD736 input < 900mV
            sngTmp0 = LOG10(sngTmp0)
            sngTmp0 = 20 * sngTmp0                'dBm (775mV @ 600Ohm = 0dBm)
            sngTmp0 = 100 * sngTmp0               'Show 2 decimals
            sngTmp0 = sngTmp0 + 0.5               'Rounding of last digt
            intTmp0 = sngTmp0
            strDis2 = str(intTmp0)
            strDis2 = Format(strDis2 , "+ 0.00")
         Else                                     'AD736 overrange
            If bitSignal = 1 Then
               strDis2 = "OvRng  "
            Else
               strDis2 = "      "
            End If
         End If
         LCD strDis2

      Case 5
         '=====================================================================
         'Relative level measurement in dB
         '=====================================================================
         If bytInitMenu = 0 Then                  'Prepare first line
            Cls
            Call ShowRange
            bytInitMenu = 1
         End If

         Call GetVoltage                          'Voltage from ADC acquisition
         Call ShowVoltage

         If bytKey_OK = 1 Then                    'Key OK pressed short
            Call Probe                            'Enable or disable 10:1 probe
         End If

         If bytKey_OK = 2 Then                    'Key OK pressed long
            If bytRelSet = 0 Then                 'Relative level was not set
               sngVoltRef = sngVolt               'Set reference voltage
               bytRelSet = 1                      'Relative level is set now
            Else                                  'Relative level was set
               bytRelSet = 0                      'Relative level is not set now
            End If
         End If
         If bytRelSet = 1 Then                    'Show relative level
            If bytRMSErr < 2 Then                 'AD736 input < 900mV
               sngTmp0 = sngVolt / sngVoltRef
               sngTmp0 = LOG10(sngTmp0)
               sngTmp0 = 20 * sngTmp0             'Relative level (dB)
               sngTmp0 = 100 * sngTmp0            'Show 2 decimals
               sngTmp0 = sngTmp0 + 0.5            'Rounding of last digt
               intTmp0 = sngTmp0
               strTmp16 = str(intTmp0)
               strTmp16 = Format(strTmp16 , "+ 0.00")
            Else                                  'AD736 overrange
               strTmp16 = "       "
            End If
            Locate 2 , 6
         Else
            Locate 2 , 7
            strTmp16 = "------"
         End If
         LCD strTmp16

      Case 6
         '=====================================================================
         'Relative level measurement in %
         '=====================================================================
         If bytInitMenu = 0 Then                  'Prepare first line
            Cls
            Call ShowRange
            bytInitMenu = 1
         End If

         Call GetVoltage                          'Voltage from ADC acquisition
         Call ShowVoltage

         If bytKey_OK = 1 Then                    'Key OK pressed short
            Call Probe                            'Enable or disable 10:1 probe
         End If

         If bytKey_OK = 2 Then                    'Key OK pressed long
            If bytRelSet = 0 Then                 'Relative level was not set
               sngVoltRef = sngVolt               'Set reference voltage
               bytRelSet = 1                      'Relative level is set now
            Else                                  'Relative level was set
               bytRelSet = 0                      'Relative level is not set now
            End If
         End If
         If bytRelSet = 1 Then                    'Show relative level
            If bytRMSErr < 2 Then                 'AD736 input < 900mV
               sngTmp0 = sngVolt / sngVoltRef
               sngTmp0 = sngTmp0 - 1
               sngTmp0 = sngTmp0 * 100            'Relative level (%)
               sngTmp1 = Abs(sngTmp0)
               If sngTmp1 < 99.999 Then           '< or > 99.999%
                  sngTmp0 = 100 * sngTmp0         'Show 2 decimals
                  intTmp0 = sngTmp0
                  strTmp16 = str(intTmp0)
                  strTmp16 = Format(strTmp16 , "+ 0.00")
               Else
                  sngTmp0 = sngTmp0 * 10          'Show 1 decimal
                  intTmp0 = sngTmp0
                  strTmp16 = str(intTmp0)
                  strTmp16 = Format(strTmp16 , "+  0.0")
               End If
            Else                                  'AD736 overrange
               strTmp16 = "       "
            End If
            Locate 2 , 6
         Else
            Locate 2 , 7
            strTmp16 = "------"
         End If
         LCD strTmp16

      Case 7
         '=====================================================================
         'Setup
         '=====================================================================
         Select Case bytMenu
            Case 71                               'Setup ADC acquisition runs
               wrdAddr = 2
               If bytInitMenu = 0 Then            'Prepare data
                  bytInitMenu = 1
                  Locate 1 , 1
                  bytTmp0 = bytMenu - 70
                  LCD "Setup " ; bytTmp0 ; "/3       "
                  ReadEEPROM bytADCRep , wrdAddr  'ADC acquisition runs
                  strUnit = " "
                  bytData = bytADCRep
                  bytDataMin = 1                  'Min repetition runs
                  bytDataMax = bytADCRepMax       'Max repetition runs
                  bytDataInc = 1                  'Increment factor
                  bytDataDec = 1                  'Decrement factor
                  bytEEPStored = 0
                  Call ClearLCDLine(2 , bytChar)
                  LCD "ADC acq runs"
                  bytCursorFirst = 14
                  Locate 2 , bytCursorFirst
               Else                               'Customize data
                  Call SetValue(2 , 1 , 1)        'Row 2, Byte, Scale 1
                  If bytEEPStored = 1 Then        'New bytADCRep was stored
                     bytADCRep = bytData          'Restore bytADCRep
                  End If
               End If

            Case 72                               'Setup ADC Reference voltage
               wrdAddr = 3
               If bytInitMenu = 0 Then            'Prepare data
                  bytInitMenu = 1
                  Locate 1 , 1
                  bytTmp0 = bytMenu - 70
                  LCD "Setup " ; bytTmp0 ; "/3       "
                  Call EEPROMReadWrd(wrdADCRef)   'ADC Reference voltage (mV)
                  strUnit = " mV"
                  wrdData = wrdADCRef
                  wrdDataMin = 20000              'Min reference voltage (mVx10)
                  wrdDataMax = 21000              'Max reference voltage (mVx10)
                  bytDataInc = 10                 'Increment factor
                  bytDataDec = 1                  'Decrement factor
                  bytEEPStored = 0
                  Call ClearLCDLine(2 , bytChar)
                  LCD "ADC ref"
                  bytCursorFirst = 9
                  Locate 2 , bytCursorFirst
               Else                               'Customize data
                  Call SetValue(2 , 2 , 1)        'Row 2, Word, Scale 1
                  If bytEEPStored = 1 Then        'New wrdADCRef was stored
                     wrdADCRef = wrdData          'Restore wrdADCRef
                     Call ADCSetCal               'Set ADC calibration factor
                  End If
               End If

            Case 73                               '2mV range offset
               wrdAddr = 5
               Select Case bytInitMenu
                  Case 0                          'Prepare data
                     bytInitMenu = 1
                     Locate 1 , 1
                     bytTmp0 = bytMenu - 70
                     LCD "Offset 2mV range"
                     Locate 2 , 1
                     LCD "Inp to GND -> OK"
                  Case 1                          'Wait for OK
                     If bytKey_OK > 0 Then
                        bytInitMenu = 2
                        Call ClearLCDLine(2 , bytChar)
                     End If
                  Case Else                       'Measure offset ADC counts
                     Call GetADCMean              'Calculate the mean ADC value
                     If bytADCResult = 1 Then     'ADC result available
                        strTmp16 = str(dwdADCMean)
                        strTmp16 = Format(strTmp16 , "     0")
                        Locate 2 , 1
                        LCD "Offset: " ; strTmp16
                     End If
                     If bytKey_OK > 0 Then        'OK pressed
                        Incr bytMenu              'Next menu
                        bytInitMenu = 0
                        If bytKey_OK = 2 Then     'OK pressed long: store to EEPROM
                           Call EEPROMWriteDwd(dwdADCMean)
                           dwdADCOffset = dwdADCMean       'Restore dwdADCOffset
                           Locate 2 , 16
                           LCD "S"
                           Wait 1
                        End If
                     End If
               End Select                         'End Select Case bytInitMenu

            Case 74                               'Setup finished
               Call ClearLCDLine(2 , bytChar)
               LCD "Setup finished"
               Wait 2
               bytMode = 1                        'Start with voltage measurement again
               bytInitMenu = 0                    'Initialize menu
               bytMenu = 10 * bytMode
               bytMenu = bytMenu + 1

         End Select                               'End Select Case bytMenu

   End Select                                     'End Select Case bytMode

Loop

End

'==============================================================================
Sub ShowRange
'Show data range in row 1 and untis in row 2...
'... and set the decimals fitting the range
'Input:  bytRange    Range number 1...6, = 2...7 if 10:1 probe is used
'        bytMode     Mode 1...5
'                    = 1 Voltage measurement
'                    = 2 Voltage  level  measurement (dBu, ref. 775mV)
'                    = 3 Voltage  level  measurement (dBV, ref. 1V)
'                    = 4 Absolute power  measurement (dBm, ref. 775mV@600Ohm)
'                    = 5 Relative level  measurement (dB, free choice ref.)
'        bytProbeSet =1: 10:1 probe is used in modes 1, 2, 3 & 5

   Cls
   If bytProbeSet = 0 Then                        '10:1 probe is not used
      strDis1 = "Range " + strRanges(bytRange) + "  FS"
   Else
      strDis1 = "Probe " + strRanges(bytRange) + "  FS"
   End If
   Locate 1 , 1
   LCD strDis1

   Select Case bytMode
      Case 1                                      'Voltage
         Locate 2 , 11
         If bytRange < 4 Then                     'Voltage in mV
            LCD "mV RMS"
         Else                                     'Voltage in V
            LCD " V RMS"
         End If
      Case 2                                      'Voltage level dBu
         Locate 2 , 14
         LCD "dBu"
      Case 3                                      'Voltage level dBV
         Locate 2 , 14
         LCD "dBV"
      Case 4                                      'Power dBm @ 600 Ohms
         Locate 2 , 14
         LCD "dBm"
      Case 5                                      'Relative level dB
         Locate 2 , 14
         LCD " dB"
      Case 6                                      'Relative level %
         Locate 2 , 14
         LCD "  %"
   End Select

   If bytMode = 1 Or bytMode = 5 Or bytMode = 6 Then       'Voltage or rel. level
      'Set number of digital places to display voltage
      Select Case bytRange
         Case 1                                   'Range 2mV
            bytDecimals = 3                       '3 decimals, mV, show x.xxx
         Case 2                                   'Range 20mV
            bytDecimals = 2                       '2 decimals, mV, show xx.xx
         Case 3                                   'Range 200mV
            bytDecimals = 1                       '1 decimal , mV, show xxx.x
         Case 4                                   'Range 2V
            bytDecimals = 3                       '3 decimals, V, show  x.xxx
         Case 5                                   'Range 20V
            bytDecimals = 2                       '2 decimals, V, show  xx.xx
         Case 6                                   'Range 200V
            bytDecimals = 1                       '1 decimal , V, show  xxx.x
         Case 7                                   'Range 200V with 10:1 probe
            bytDecimals = 1                       '1 decimal , V, show  xxx.x
      End Select
   End If

End Sub

'==============================================================================
Sub Probe
'Enable or disable 10:1 probe
'Correct current range display
'Input:  bytProbeSet current probe status, =0: probe is not used actually
'        bytRange    current range
'Output: bytProbeSet changed probe status, 0->1 or 1->0
'        bytRange    changed range, increased or decreased
'        bytRangeMin corrected min. range
'        bytRangeMax corrected max. range

   If bytProbeSet = 0 Then                        '10:1 probe was not used
      bytProbeSet = 1                             '10:1 probe is used now
      bytRange = bytRange + 1
      bytRangeMin = 2
      bytRangeMax = 7
   Else                                           '10:1 probe was used
      bytProbeSet = 0                             '10:1 probe is not used now
      bytRange = bytRange - 1
      bytRangeMin = 1
      bytRangeMax = 6
   End If
   bytProbeCur = bytProbeSet                      'Save current probe status
   Call ShowRange                                 'Correct range display

End Sub

'==============================================================================
Sub GetVoltage
'Get ADC counts and calculate ADC input voltage with 18bit calibration factor
'Input voltage to RMS chip AD736 = ADC input is always 200mV
'According to selected ranges input is multiplied and/or divided
'Range 1     2mV     gain 100       to give 200mV
'Range 2    20mV     gain  10       to give 200mV
'Range 3   200mV     gain   1       to give 200mV
'Range 4     2 V     gain   1/10    to give 200mV
'Range 5    20 V     gain   1/100   to give 200mV
'Range 6   200 V     gain   1/1000  to give 200mV
'Input:  dwdADCMean  ADC counts mean value
'        bytMode     Current mode 1...5
'                    = 1 Voltage measurement
'                    = 2 Voltage  level  measurement (dBu, ref. 775mV)
'                    = 3 Voltage  level  measurement (dBV, ref. 1V)
'                    = 4 Absolute power  measurement (dBm, ref. 775mV@600Ohm)
'                    = 5 Relative level  measurement (dB, free choice ref.)
'        bytProbeSet =1: 10:1 probe is used
'Output: sngVolt     Voltage measured by ADC (mV). Is not meter input voltage!
'        bytRMSErr   Indicator for AD736 RMS input
'                    = 0: is <= 200mV (within range for in spec precision)
'                    = 1: is >  200mV, but < 900mV: still OK, but worse precision
'                    = 2: is >  900mV (out of range)

   'dwdADCMean = 0
   Call GetADCMean                                'Calculate the mean ADC value
   If bytADCResult = 1 Then                       'ADC result available
      If bytRange = 1 Then                        'Range 2mV
         If dwdADCMean > dwdADCOffset Then
            dwdADCMean = dwdADCMean - dwdADCOffset       'Offset correction
         Else
            dwdADCMean = 0
         End If
      End If
      If dwdADCMean = 0 Then dwdADCMean = 1       'Trouble with LOG10 if = 0

      dblTmp0 = dwdADCMean
      dblTmp0 = dblTmp0 * dblADCCalFac            'mV calibration
      bytRMSErr = 0                               'AD736 RMS input is OK
      If dblTmp0 > 200 Then                       'RMS input is > 200mV
         bytRMSErr = 1
         If dblTmp0 > 900 Then                    'RMS input is > 900mV
            bytRMSErr = 2
         End If
      End If
      If bytRMSErr > 0 Then                       '> 200mV (AD736 conversion limit)
         Locate 1 , 13
         If bitSignal = 1 Then
            LCD "**"
         Else
            LCD "  "
         End If
      Else
         Locate 1 , 13
         LCD "  "
      End If

      Select Case bytMode
         Case 2 To 4                              'dBu, dBV or dBm
            'Multiply voltage (mV) with powers of ten
            'according to gain or attenuation of selected range

            Select Case bytRange
               Case 1                             'Range 2mV, multiplier 1/100
                  dblTmp0 = 0.01 * dblTmp0
                  sngVolt = dblTmp0
               Case 2                             'Range 20mV, multiplier 1/10
                  dblTmp0 = 0.1 * dblTmp0
                  sngVolt = dblTmp0
               Case 3                             'Range 200mV, no multiplier
                  sngVolt = dblTmp0
               Case 4                             'Range 2V, multiplier 1/100
                  dblTmp0 = 0.01 * dblTmp0
                  sngVolt = dblTmp0
               Case 5                             'Range 20V, multiplier 1/10
                  dblTmp0 = 0.1 * dblTmp0
                  sngVolt = dblTmp0
               Case 6                             'Range 200V, no multiplier
                  sngVolt = dblTmp0
               Case 7                             'Range 200V with 10:1 probe
                  dblTmp0 = 10.0 * dblTmp0
                  sngVolt = dblTmp0
            End Select                            'End Select bytRange

         Case Else                                'bytMode=1 or 5 (voltage, rel. level)
            If bytRange = 7 Then                  'Range 200V with 10:1 probe
               dblTmp0 = 10.0 * dblTmp0
               sngVolt = dblTmp0
            Else
               sngVolt = dblTmp0                  'Voltage at ADC input (mV)
            End If
      End Select                                  'End Select bytMode

   End If

End Sub

'==============================================================================
Sub ShowVoltage
'Show formatted voltage in mV or V in row 2
'Range  2mV    show x.xxx
'Range  20mV   show xx.xx
'Range  200mV  show xxx.x
'Range  2V     show x.xxx
'Range  20V    show xx.xx
'Range  200V   show xxx.x
'Range 2000V   show xxx.x     (200V range with 10:1 probe only)
'Input:        sngVolt        Voltage measured by ADC (mV),
'                             not meter input voltage!
'              bytDecimals    Number of decimals (0...3)

   Locate 2 , 1
   sngTmp0 = 10 * sngVolt                         'one more digit before fix
   sngTmp0 = sngTmp0 + 0.5                        'Rounding of last digt
   dwdTmp0 = sngTmp0                              'Fix
   strDis2 = str(dwdTmp0)
   If bytRMSErr < 2 Then                          'AD736 input < 900mV
      Select Case bytDecimals
         Case 1                                   '1 decimal
            strDis2 = Format(strDis2 , "  0.0")
         Case 2                                   '2 decimals
            strDis2 = Format(strDis2 , " 0.00")
         Case 3                                   '3 decimals
            strDis2 = Format(strDis2 , "0.000")
      End Select
      strDis2 = strDis2 + " "
   Else                                           'AD736 input >= 900mV
      If bitSignal = 1 Then
         strDis2 = "OvRng "
      Else
         strDis2 = "      "
      End If
   End If
   LCD strDis2

End Sub

'==============================================================================
Timer0_isr:
'Interrupt Service Routine for Timer0 (Keys)
'got from Thomas Baer in microcontroller.net
'Original (ASM and C) by Peter Dannegger
'http://www.mikrocontroller.net/topic/tasten-entprellen-bulletproof
'Input:  Key_port       port connected to the keys
'        bytKey_mask    Bit pattern to mask the keys enabled for short/long
'                       Key number (0...7) to be checked for short/long
'                       e.g. "00000101" Keys 0 and 2 are short/long keys
'Output: bytKey_press   bit pattern, key which changed from open to pressed
'                       Reset after evaluation in the main program
'                       Bits remain = 1 until resetted in the main program
'        bytKey_state   bit pattern, Key state, 1:pressed or 0:not pressed
'                       Bits = 1 while key is pressed, = 0: Key is released
'        bytKey_save    last state of bytKey_state before entering Timer0_isr
'                       Bits 0...7: keys 0...7 @ Key_port

   Timer0 = Presettimer0
   bytKey_save = bytKey_state                     'save old status before keys were pressed
   bytIwr0 = Key_port                             'get Bits of the KeyPort
   bytIwr0 = Not bytIwr0
   bytIwr0 = bytIwr0 Xor bytKey_state
   bytKey_ct0 = bytKey_ct0 And bytIwr0
   bytKey_ct1 = bytKey_ct1 And bytIwr0
   bytKey_ct0 = Not bytKey_ct0
   bytKey_ct1 = bytKey_ct1 Xor bytKey_ct0
   bytIwr0 = bytIwr0 And bytKey_ct0
   bytIwr0 = bytIwr0 And bytKey_ct1
   bytKey_state = bytKey_state Xor bytIwr0
   bytIwr0 = bytIwr0 And bytKey_state
   bytKey_press = bytKey_press Or bytIwr0

   'Key long press or auto repeat
   bytIwr0 = bytKey_state AND bytKey_mask         'Mask out short/long keys
   If bytIwr0 > 0 Then                            'Short/long key still pressed
      Incr bytTimer0_keys                         'Key press counter
   End If

   'ADC high level signal
   If bytTimer0_Signal < 70 Then
      Incr bytTimer0_Signal
   Else
      Toggle bitSignal
      bytTimer0_Signal = 0
   End If

Return

'=========================================================================Keys=
Function KeyPressedLong(byVal bytKey_Long As Byte) As Byte
   'Check if key bytKey_Long (0...7) in Key_port is pressed short or long
   'Output: KeyPressedLong =  0: evaluating, key is still pressed
   '                          1: key is released, pressed short
   '                          2: key is released, pressed long

   If bytKey_state.bytKey_Long = 1 Then           'Key still pressed
      KeyPressedLong = 0
   Else                                           'Key released
      If bytTimer0_keys < bytTimer0_long Then     'short time elapsed
         KeyPressedLong = 1                       'Pressed short
      Else
         KeyPressedLong = 2                       'Pressed long
      End If
      bytTimer0_keys = 0
      bytKey_press.bytKey_Long = 0                'Reset pressed flag
   End If
End Function

'=========================================================================Keys=
Function KeyAutoRepeat(byVal bytKey_Rep As Byte) As Byte
'Check if key bytKey_Rep (0...7) in Key_port is pressed long for auto repeat
'Key status bytKey_state.bytKey_Rep is returned from Timer0_isr
'Input:  bytKey_state.bytKey_Rep =0: Key #bytKey_Rep is not   pressed
'                                =1: Key #bytKey_Rep is still pressed
'        bytTimer0_keys          Counter returned from Timer0_isr
'                                Incremented per 12.8 msec
'                                For different Timer0 timing check conditions
'                                "bytTimer0_keys >/< 20"
'Output: KeyAutoRepeat           =0: key is not pressed
'                                =1: key is pressed short, no auto repeat
'                                =2: key is pressed long,     auto repeat

   KeyAutoRepeat = 0
   If bytKey_state.bytKey_Rep = 1 Then            'Key still pressed
      If bytTimer0_keys > 20 Then
         KeyAutoRepeat = 2                        'Pressed long, auto repeat
         bytTimer0_keys = 0                       'Reset Timer0 counter
         Waitms 100                               'Step speed
      End If
   Else                                           'Key released
      If bytTimer0_keys < 20 Then                 '~ 250 msec
         KeyAutoRepeat = 1                        'Pressed short, no auto repeat
         bytTimer0_keys = 0                       'Reset Timer0 counter
         bytKey_press.bytKey_Rep = 0              'Reset pressed flag
      End If
   End If
End Function

'==============================================================================
Sub SetValue(byVal bytRow As Byte , byVal bytType As Byte , byVal wrdScale As Word)
'Set EEPROM value via keys and restore it to EEPROM
'EEPROM data are byte, Word or Integer, perhaps scaled with factors 10, 100 or 1000
'e.g. 1.234V ist stored as 1234mV in EEPROM
'Values are changed via up and down key
'Input:  bytRow                  LCD row to display value
'        bytType                 Type of value
'                                1 = Byte
'                                2 = Word
'        bytData                 Input variable for bytType = 1 (Byte)
'        wrdData                 Input variable for bytType = 2 (Word)
'        wrdScale                Scale factor of EEPROM value, 1, 10, 100 or 1000
'                                EEPROM value = real value * wrdScale
'                                e.g. Word variable (mV) to be shown as Volt:
'                                wrdScale=1000, 1234 mV divided by 1000, shown as 1.234V
'                                stored as 1.234 * 1000 (mV, Word)
'        bytCursorFirst          First LCD column to show data
'        wrdAddr                 EEPROM address to store value

   Select Case bytType
      Case 1                                      'Input byte variable bytData
         strTmp16 = Str(bytData)
      Case 2                                      'Input Word variable wrdData
         strTmp16 = Str(wrdData)
   End Select

   Select Case wrdScale
      Case 1
         strTmp16 = Format(strTmp16 , " 0")
      Case 10
         strTmp16 = Format(strTmp16 , " 0.0")
      Case 100
         strTmp16 = Format(strTmp16 , " 0.00")
      Case 1000
         strTmp16 = Format(strTmp16 , " 0.000")
   End Select
   strTmp16 = strTmp16 + strUnit
   Locate bytRow , bytCursorFirst
   LCD strTmp16

   'Up key  pressed short: increment by bytDataInc
   '        pressed long : with autorepeat
   If bytkey_Up > 0 Then                          'Up short or long/autorepeat
      Select Case bytType
         Case 1
            bytData = bytData + bytDataInc
            If bytData > bytDataMax Then bytData = bytDataMax
         Case 2
            wrdData = wrdData + bytDataInc
            If wrdData > wrdDataMax Then wrdData = wrdDataMax
      End Select

   End If

   'Both Up and OK key pressed: return to previous menu
   If bytkey_Down > 0 And bytKey_OK > 0 Then
      If bytMenu > 21 Then decr bytMenu
      bytInitMenu = 0
      Exit Sub
   End If

   'Down key   pressed short: decrement by 1
   '           pressed long : with autorepeat
   If bytkey_Down > 0 Then                        'Down short or long/autorepeat
      Select Case bytType
         Case 1
            If bytData >= bytDataDec Then
               bytData = bytData - bytDataDec
               If bytData < bytDataMin Then bytData = bytDataMin
            End If
         Case 2
            If wrdData >= bytDataDec Then
               wrdData = wrdData - bytDataDec
               If wrdData < wrdDataMin Then wrdData = wrdDataMin
            End If
      End Select
   End If

   'OK key pressed short: next menu, data are not stored
   '        pressed long : store current data to EEPROM, next menu
   Select Case bytKey_OK
      Case 1                                      'OK short, finish
         Incr bytMenu                             'Next menu
         bytInitMenu = 0
      Case 2                                      'OK long, store & finish
         Select Case bytType
            Case 1                                'Input byte variable bytData
               WriteEEPROM bytData , wrdAddr
            Case 2                                'Input Word variable wrdData
               Call EEPROMWriteWrd(wrdData)
         End Select
         Locate 2 , 16
         LCD "S"
         bytEEPStored = 1                         'Data stored
         Incr bytMenu                             'Next menu
         bytInitMenu = 0
         Wait 1
   End Select

End Sub

'==============================================================================
Sub ClearLCDLine(byVal bytRow As Byte , byVal bytChars As Byte)
'Clear LCD bytChars characters in line bytRow
   Locate bytRow , 1
   LCD SPC(bytChars)
   Locate bytRow , 1
End Sub

'==========================================================================ADC=
Sub GetADCMean
'Wait for external ADC to finish a single conversion
'calculate the mean value of bytADCRep single conversions
'Input:  bytADCRes      ADC resolution (18 bits)
'        bytADCRep      Number of repetitive runs for mean value
'        dwdADCOvl      ADC overflow count for chosen ADC resolution
'Output: dwdADCMean     ADC Mean value, running average
'        bytADCOvl      =1: at least one ADC overflow detected, =0 else
'        bytADCResult   =1: ADC mean available

   bytADCOvl = 0                                  'ADC value within range
   bytADCResult = 0

   'Read MCP3421 ADC conversion result
   'Data bytes  bytADCVal(1...4)are overlayed with lngADCVal (DWord)
   'For 18 bit resolution bytes 1, 2 and 3 are read
   I2CStart
   I2CwByte bytADCRead                            'Read address
   I2CrByte bytADCVal(3) , Ack                    'Data byte #3 for 18 bit resolution
   I2CrByte bytADCVal(2) , Ack                    'Data byte #2
   I2CrByte bytADCVal(1) , Ack                    'Data byte #1
   I2CrByte bytADCConf , Nack                     'Configuration byte
   I2CStop

   'Check RDY-bit (#7) of the configuration byte
   If bytADCConf.7 = 0 Then                       'RDY=0 if conversion is complete
      'Valid for 18 bit resolution:
      'bit #1 of bytADCVal(3) is the sign bit, 0=pos., 1=neg.
      'Set lngADCVal = 0 if negative, possible voltage ripple < 0
      bytADCVal(4) = 0                            'Complete not read byte 4
      If bytADCVal(3).1 = 1 Then                  '"-" Sign bit in ADC digital code
         lngADCVal = 0                            'negative, set = 0
      End If
      If lngADCVal >= dwdADCOvl Then              'ADC overflow
         bytADCOvl = 1                            'ADC value out of range
      End If
      If bytADCRep = 1 Then                       'One shot
         dwdADCMean = lngADCVal
      Else
         'Running average
         'Current ADC value is summed in dwdADCSum and stored in array lngADCArr
         'With a new ADC value the last value is substracted from dwdADCSum
         'and then added to dwdADCSum
         'If repetitions reached end value bytADCRep the array index begins with 1 again
         'The average is calculated with every new ADC value
         dwdADCSum = dwdADCSum - lngADCArr(bytADCRun)       'Substract last ADC value
         dwdADCSum = dwdADCSum + lngADCVal        'Add new ADC value
         lngADCArr(bytADCRun) = lngADCVal         'New value into array
         Incr bytADCRun                           'Next array index
         If bytADCRun >= bytADCRep Then           'Repetitions complete
            bytADCRun = 1
         End If
         bytTmp0 = bytADCRep - 1
         dwdADCMean = dwdADCSum / bytTmp0         'Mean value
      End If
      bytADCResult = 1
   End If

End Sub

'==========================================================================ADC=
Function ADCCheck() As Byte
'Check I2C-ADC, try th write configuration byte bytADCWrite to the device
'CheckI2C = 1: OK, 0: else

   I2CInit
   I2CStart
   I2CwByte bytADCWrite                           'Try to write Write address
   I2CStop
   If Err = 1 Then                                'Error raised
      ADCCheck = 0                                'Not OK
   Else
      ADCCheck = 1                                'Is OK
   End If
End Function

'==========================================================================ADC=
Sub ADCSetCal
'Set ADC MCP3421 calibration factor for 18bit resolution
'Input:  wrdADCRef      Reference voltage (mVx10), nominal 2048 mV
'Output: dblADCCalFac   mV Calibration factor 18bit

   dblADCCalFac = wrdADCRef                       'As mV x 10
   dblADCCalFac = 0.2 * dblADCCalFac              'Now mV
   dblADCCalFac = dblADCCalFac / 262144.0         'mV ADC calibration factor 18Bit

End Sub

'==========================================================================ADC=
Sub ADCSetGain
'Send configuration byte to ADC MCP3421 configuration register
'to set Programmable Gain Amplifier (PGA) gain 1, 2, 4 or 8
'and to set samples per second (SPS) = Resolution 16 or 18bit
'Conversion Mode: Continuous
'Sample Rate:   3.75 SPS (18 bits), variable bytADCRes
'... or        15    SPS (16 bits)
'bit  name  function
'7    RDY   Ready bit indicates the state of conversion
'           Read mode:
'           RDY=1: Output register has not been updated
'           RDY=0: Output register has been updated with the latest conversion data
'           Write mode:
'           Continuous conversion: No effect
'           One-Shot conversion mode:
'           RDY=1: New conversion
'           RDY=0: No effect
'6,5  C1,C0 Channel selection bits, not used (=00) in MCP3421 (1 channel only)
'4    O/C   Conversion Mode Bit
'           O/C=1: Continuous Conversion Mode
'           O/C=0: One-Shot Conversion Mode
'3,2  S1,S0 Sample Rate Selection Bit (SPS=samples per second)
'           00 =  240 SPS (12 bits)
'           01 =   60 SPS (14 bits)
'           10 =   15 SPS (16 bits)
'           11 = 3.75 SPS (18 bits)
'1,0  G1,G0 Programmable Gain Amplifier (PGA) Selection bit
'           00= 1 V/V
'           01= 2 V/V
'           10= 4 V/V
'           11= 8 V/V
'Input:  bytADCGain  Gain select:1/2/4/8
'Output: ./.

   Select Case bytADCGain                         'Set ADC configuration bytes
      Case 1                                      'PGA Gain 1
         bytADCConf = &B10011100                  'PGA Gain 1/18 bit
      Case 2
         bytADCConf = &B10011101                  'PGA Gain 2/18 bit
      Case 4
         bytADCConf = &B10011110                  'PGA Gain 4/18 bit
      Case 8
         bytADCConf = &B10011111                  'PGA Gain 8/18 bit
   End Select

   I2CStart
   I2CwByte bytADCWrite                           'Write address
   I2CwByte bytADCConf                            'Write configuration byte
   I2CStop

   Call ClearLCDLine(2 , 16)                      'Not OK
   If Err = 1 Then                                'Error raised
      LCD "Err Set ADC Gain"
   Else
      LCD "Gain " ; bytADCGain ; " / " ; bytADCRes ; "bit"       'Is OK
   End If
   Wait 1
End Sub

'=======================================================================EEPROM=
Sub CheckEEPROM
'Read voltmeter data from EEPROM or initialize data
'In a new controller or erased EEPROM data = 255
'If voltmeter data are stored, byte # 1 = 1
'Voltmeter data:
'Byte Variable          Type     Description
' 0   --                Byte     Not used
' 1   --                Byte     = 1: data are stored
' 2   bytADCRep         Byte     Repetitive ADC runs for mean value
' 3   wrdADCRef         Word     ADC reference voltage (mV)

   ReadEeprom bytTmp1 , 1                         'Check Address #1
   Locate 1 , 1
   If bytTmp1 = 1 Then                            'Read available EEPROM data
      'LCD "Read EEPROM...  "
      ReadEEPROM bytADCRep , 2                    'Repetitive ADC runs data accumulation
      wrdAddr = 3
      Call EEPROMReadWrd(wrdADCRef)               'ADC reference voltage (mV)
      wrdAddr = 5
      Call EEPROMReadDwd(dwdADCOffset)            '2mV ADC offset counts
   Else
      LCD "Init EEPROM...  "
      bytADCRep = 1                               'Repetitive ADC runs data accumulation
      WriteEEPROM bytADCRep , 2
      wrdAddr = 3
      wrdADCRef = 20480                           '2048mV x 10
      Call EEPROMWriteWrd(wrdADCRef)              'ADC reference voltage (mVx10)
      wrdAddr = 5
      dwdADCOffset = 0
      Call EEPROMWriteDwd(dwdADCOffset)           '2mV ADC offset counts
      bytTmp1 = 1
      WriteEEPROM bytTmp1 , 1                     'EEPROM initialized
      Locate 2 , 1
      LCD "... Init done   "                      '
      Wait 2
   End If

End Sub

'=======================================================================EEPROM=
Sub EEPROMWriteWrd(byRef wrdVal As Word)
   'Decode wrdWord to 2 bytes of bytWrdEEP for writing to EEPROM
   'Input:  wrdWord to be decoded & written to EEPROM
   '        wrdAddr first EEPROM address

   wrdEEPROM = wrdVal                             'wrdEEPROM overlayed @ bytWrdEEP
   For bytTmp0 = 1 To 2
      WriteEEPROM bytWrdEEP(bytTmp0) , wrdAddr
      Incr wrdAddr
   Next bytTmp0

End Sub

'=======================================================================EEPROM=
Sub EEPROMReadWrd(byRef wrdVal As Word)
   'Encode 2 bytes read from EEPROM to wrdWord
   'Input:  wrdAddr first EEPROM address
   'Output: wrdWord encoded unsigned integer (word) variable

   For bytTmp0 = 1 To 2
      ReadEEPROM bytWrdEEP(bytTmp0) , wrdAddr
      Incr wrdAddr
   Next bytTmp0
   wrdVal = wrdEEPROM                             'wrdEEPROM overlayed @ bytWrdEEP

End Sub

'=======================================================================EEPROM=
Sub EEPROMWriteDwd(byRef dwdVar As DWord)
   'Decode dwdVar to 4 bytes of bytDwdEEP for writing to EEPROM
   'Input:  dwdVar to be decoded & written to EEPROM (4 bytes)
   '        wrdAddr first EEPROM address
   Local i As Byte

   dwdEEPROM = dwdVar
   For i = 1 To 4
      WriteEEPROM bytDwdEEP(i) , wrdAddr
      Incr wrdAddr
   Next i
End Sub

'=======================================================================EEPROM=
Sub EEPROMReadDwd(byRef dwdVar As DWord)
   'Encode 4 bytes read from EEPROM to dwdVar
   'Input:  wrdAddr first EEPROM address
   'Output: dwdVar encoded unsigned double integer (DWord) variable
   Local i As Byte

   For i = 1 To 4
      ReadEEPROM bytDwdEEP(i) , wrdAddr
      Incr wrdAddr
   Next i
   dwdVar = dwdEEPROM                             'dwdEEPROM overlayed @ bytDwdEEP
End Sub